#! /usr/bin/env python
# -*- encoding: utf-8 -*-

from __future__ import absolute_import, print_function

import argparse
import base64
import hashlib
import io
import json
import os
import platform
import queue
import signal
import socket
import subprocess
import sys
import time
import traceback
import uuid
import webbrowser
# `pip install futures` for python2
from concurrent.futures import ThreadPoolExecutor
from subprocess import PIPE

import requests
import six
import tornado.escape
import tornado.ioloop
import tornado.web
import tornado.websocket
from logzero import logger
from tornado import gen
from tornado.concurrent import run_on_executor
from tornado.escape import json_encode
from tornado.log import enable_pretty_logging

from .web.handlers.page import (
    BaseHandler, BuildWSHandler, DeviceCodeDebugHandler, DeviceConnectHandler,
    DeviceHierarchyHandler, DeviceHierarchyHandlerV2, DeviceScreenshotHandler,
    DeviceWidgetListHandler, MainHandler, VersionHandler, WidgetPreviewHandler)
from .web.handlers.proxy import StaticProxyHandler
from .web.utils import current_ip, tostr
from .web.version import __version__

enable_pretty_logging()

__dir__ = os.path.dirname(os.path.abspath(__file__))

is_closing = False


def signal_handler(signum, frame):
    global is_closing
    print('exiting...')
    is_closing = True


def stop_server():
    tornado.ioloop.IOLoop.instance().stop()


def try_exit():
    global is_closing
    if is_closing:  # clean up here
        stop_server()
        logger.info('exit success')


class QuitHandler(BaseHandler):
    def get(self):
        stop_server()
        self.write({"success": True, "description": "Successfully quited"})


class CropHandler(BaseHandler):
    def get(self):
        """ used for crop image """
        pass


def make_app(settings={}):
    application = tornado.web.Application(
        [
            (r"/", MainHandler),
            (r"/api/v1/version", VersionHandler),
            (r"/api/v1/connect", DeviceConnectHandler),
            (r"/api/v1/crop", CropHandler),
            (r"/api/v1/devices/([^/]+)/screenshot", DeviceScreenshotHandler),
            (r"/api/v1/devices/([^/]+)/hierarchy", DeviceHierarchyHandler),
            (r"/api/v1/devices/([^/]+)/exec", DeviceCodeDebugHandler),
            (r"/api/v1/devices/([^/]+)/widget", DeviceWidgetListHandler),
            (r"/api/v1/widgets", DeviceWidgetListHandler),  # add widget
            (r"/api/v1/widgets/([^/]+)", DeviceWidgetListHandler),
            # v2
            (r"/api/v2/devices/([^/]+)/hierarchy", DeviceHierarchyHandlerV2),
            # widgets
            (r"/widgets/([^/]+)", WidgetPreviewHandler),
            (r"/widgets/(.+/.+)", tornado.web.StaticFileHandler, {
                "path": "./widgets"
            }),
            # cache static assets
            (r"/(unpkg.com/.*)", StaticProxyHandler),
            (r"/(cdn.jsdelivr.net/.*)", StaticProxyHandler),
            (r"/ws/v1/build", BuildWSHandler),
            (r"/quit", QuitHandler),
        ],
        **settings)
    return application


def get_running_version(addr: str):
    """
    Returns:
        None if not running
        version string if running
    """
    try:
        r = requests.get(f"{addr}/api/v1/version",
                         timeout=2.0)
        if r.status_code == 200:
            return r.json().get("version", "dev")
    except requests.exceptions.ConnectionError:
        pass
    except Exception as e:
        print("Unknown error: %r" % e)


def run_web(debug=False, port=17310, open_browser=False, force_quit=False):
    version = get_running_version(f"http://localhost:{port}")
    if version:
        if force_quit:
            logger.info(f"quit previous weditor server (version: {version})")
            requests.get(f"http://localhost:{port}/quit")
            time.sleep(.5)
        else:
            sys.exit(f"Another weditor({version}) is already running")

    if open_browser:
        # webbrowser.open(url, new=2)
        webbrowser.open(f'http://localhost:{port}', new=2)

    application = make_app({
        'static_path': os.path.join(__dir__, 'static'),
        'template_path': os.path.join(__dir__, 'templates'),
        'debug': debug,
    })
    print('listening on http://%s:%d' % (current_ip(), port))
    if debug:
        logger.info("enable debug mode")
    signal.signal(signal.SIGINT, signal_handler)
    application.listen(port)
    tornado.ioloop.PeriodicCallback(try_exit, 100).start()
    # tornado.ioloop.IOLoop.instance().add_callback(consume_queue)
    tornado.ioloop.IOLoop.instance().start()


def create_shortcut():
    if os.name != 'nt':
        sys.exit("Only valid in Windows")

    import pythoncom  # pyint: disable=import-error
    from win32com.shell import shell
    from win32com.shell import shellcon
    # Refs
    # - https://github.com/pearu/iocbio/blob/master/installer/utils.py
    # - https://blog.csdn.net/thundor/article/details/5968581
    ilist = shell.SHGetSpecialFolderLocation(0, shellcon.CSIDL_DESKTOP)
    dtpath = shell.SHGetPathFromIDList(ilist).decode('utf-8')

    shortcut = pythoncom.CoCreateInstance(shell.CLSID_ShellLink, None,
                                          pythoncom.CLSCTX_INPROC_SERVER,
                                          shell.IID_IShellLink)
    launch_path = sys.executable
    shortcut.SetPath(launch_path)
    shortcut.SetArguments("-m weditor")
    shortcut.SetDescription(launch_path)
    shortcut.SetIconLocation(sys.executable, 0)
    shortcut.QueryInterface(pythoncom.IID_IPersistFile).Save(
        dtpath + "\\WEditor.lnk", 0)
    print("Shortcut created. " + dtpath + "\\WEditor.lnk")


def main():
    # yapf: disable
    ap = argparse.ArgumentParser(
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    ap.add_argument("-v", "--version", action="store_true", help="show version")
    ap.add_argument('-q', '--quiet', action='store_true', help='quite mode, no open new browser')
    ap.add_argument('-d', '--debug', action='store_true', help='open debug mode')
    ap.add_argument('-p', '--port', type=int, default=17310, help='local listen port for weditor')
    ap.add_argument("-f", "--force-quit", action='store_true', help="force quit before start")
    ap.add_argument('--shortcut', action='store_true', help='create shortcut in desktop')
    args = ap.parse_args()
    # yapf: enable

    if args.version:
        print(__version__)
        return

    if args.shortcut:
        create_shortcut()
        return

    if sys.platform == 'win32' and sys.version_info[:2] == (3, 8):
        import asyncio
        asyncio.set_event_loop_policy(asyncio.WindowsSelectorEventLoopPolicy())

    open_browser = not args.quiet and not args.debug
    print(args.quiet, args.debug, open_browser)
    run_web(args.debug, args.port, open_browser, args.force_quit)


if __name__ == '__main__':
    main()



# coding: utf-8
#

import abc
from PIL import Image
from . import uidumplib


class DeviceMeta(metaclass=abc.ABCMeta):
    def screenshot(self) -> Image.Image:
        pass

    def dump_hierarchy(self) -> str:
        pass
    
    @abc.abstractproperty
    def device(self):
        pass


class _AndroidDevice(DeviceMeta):
    def __init__(self, device_url):
        import uiautomator2 as u2
        d = u2.connect(device_url)
        self._d = d

    def screenshot(self):
        return self._d.screenshot()

    def dump_hierarchy(self):
        return uidumplib.get_android_hierarchy(self._d)
    
    def dump_hierarchy2(self):
        current = self._d.app_current()
        page_xml = self._d.dump_hierarchy(pretty=True)
        page_json = uidumplib.android_hierarchy_to_json(page_xml.encode('utf-8'))
        return {
            "xmlHierarchy": page_xml, 
            "jsonHierarchy": page_json,
            "activity": current['activity'],
            "packageName": current['package'],
            "windowSize": self._d.window_size(),
        }

    @property
    def device(self):
        return self._d


class _AppleDevice(DeviceMeta):
    def __init__(self, device_url):
        import wda
        c = wda.Client(device_url)
        self._client = c
        self.__scale = c.session().scale

    def screenshot(self):
        return self._client.screenshot(format='pillow')

    def dump_hierarchy(self):
        return uidumplib.get_ios_hierarchy(self._client, self.__scale)
    
    def dump_hierarchy2(self):
        return {
            "jsonHierarchy": uidumplib.get_ios_hierarchy(self._client, self.__scale),
            "windowSize": self._client.session().window_size(),
        }

    @property
    def device(self):
        return self._client.session()


class _GameDevice(DeviceMeta):
    def __init__(self, device_url):
        import neco
        d = neco.connect(device_url)
        self._d = d

    def screenshot(self):
        return self._d.screenshot()

    def dump_hierarchy(self):
        return self._d.dump_hierarchy()

    @property
    def device(self):
        return self._d


cached_devices = {}


def connect_device(platform, device_url):
    """
    Returns:
        deviceId (string)
    """
    device_id = platform + ":" + device_url
    if platform == 'android':
        d = _AndroidDevice(device_url)
    elif platform == 'ios':
        d = _AppleDevice(device_url)
    else:
        d = _GameDevice(device_url or "localhost")

    cached_devices[device_id] = d
    return device_id


def get_device(id):
    d = cached_devices.get(id)
    if d is None:
        platform, uri = id.split(":", maxsplit=1)
        connect_device(platform, uri)
    return cached_devices[id]



# coding: utf-8
#

import uuid
import json
import os
import signal
import subprocess
import sys
from subprocess import PIPE
from typing import Union

from logzero import logger


class ConsoleKernel():
    __instance = {}

    @classmethod
    def get_singleton(cls, name='default'):
        if cls.__instance.get(name) is None:
            logger.info("create singleton process")
            cls.__instance[name] = ConsoleKernel()
        return cls.__instance[name]

    def __init__(self):
        self._p = None
        self._magic = "M:" + str(uuid.uuid4()).replace("-", "") + "=="
        self._start_jsonrpc_server()

    def _start_jsonrpc_server(self):
        curdir = os.path.dirname(os.path.abspath(__file__))
        rpcserver_path = os.path.join(curdir, "./jsonrpc_server.py")
        self._p = subprocess.Popen([sys.executable, "-u", rpcserver_path, "--magic", self._magic],
                                   bufsize=1,
                                   stdin=PIPE,
                                   stdout=PIPE,
                                   stderr=subprocess.STDOUT,
                                   universal_newlines=True) # yapf: disable
        data, finished = self.read_response()
        assert finished == True, 'jsonrpc_server started failed'

    def send_interrupt(self):
        if not self.is_closed():
            self._p.send_signal(signal.SIGINT)

    def is_closed(self):
        return self._p.poll() is not None

    def read_response(self) -> tuple:
        """
        Returns:
            tuple: (output, finished: bool)
        """
        
        line = self._p.stdout.readline()
        logger.info("READ: %r", line)
        if self.is_closed():
            self._p = None
            return line, True
        if line.rstrip().endswith(self._magic):
            return line.rstrip()[:-len(self._magic)], True
        return line, False

    def read_until_finished(self):
        output = ""
        while True:
            data, finished = self.read_response()
            output += data
            if finished:
                break
        return output

    def call(self, method: str, params: Union[list, dict]):
        send_data = json.dumps({
            "jsonrpc": "2.0",
            "method": method,
            "params": params
        })
        logger.info("SEND: %s", send_data)
        self._p.stdin.write(send_data + "\n")

    def call_output(self, method: str, params: Union[list, dict]) -> str:
        self.call(method, params)
        return self.read_until_finished()


if __name__ == "__main__":
    s = ConsoleKernel.get_singleton()
    s = ConsoleKernel.get_singleton()
    s.call("pow", [2, 5])
    output = s.read_until_finished()
    print("Output:", output)
    s.call("run_device_code", ["android:", "d.info"])
    output = s.read_until_finished()
    print("Output:", output)

    output = s.call_output("run_device_code", ["android:", "d.info"])
    print("Output:", output)


# coding: utf-8
#

import hashlib
import os
import re

import tornado.httpclient
import tornado.web
from logzero import logger

from .page import BaseHandler


class StaticProxyHandler(tornado.web.StaticFileHandler):
    http_client = tornado.httpclient.AsyncHTTPClient()

    def initialize(self, path: str = None, default_filename: str = None) -> None:
        self.root = path if path else os.path.expanduser("~")
        self.default_filename = default_filename

    async def download_file(self, path: str) -> str:
        """
        Returns:
            download file path
        
        Raises:
            tornado.web.HTTPError
        """
        cache_path = os.path.join(self.settings.get("static_path"), "cdn_libraries", path)
        if os.path.exists(cache_path):
            return cache_path
        
        # cache to local directory
        local_cache_dir = os.path.expanduser("~/.weditor/cache")
        cache_path = os.path.join(local_cache_dir, path)
        if os.path.exists(cache_path):
            return cache_path
        
        os.makedirs(os.path.dirname(cache_path), exist_ok=True)

        request = tornado.httpclient.HTTPRequest(
            url="https://"+path,
            method="GET",
            validate_cert=False  # fix certificate validate error
        )

        response = await self.http_client.fetch(request, raise_error=False)
        if response.code != 200:
            raise tornado.web.HTTPError(404)
        
        with open(cache_path, 'wb') as f:
            f.write(response.body)
        
        return cache_path

    async def get(self, path: str, include_body: bool = True) -> None:
        abspath = await self.download_file(path)
        await super().get(abspath, include_body)



# coding: utf-8
#
# WRT:{quoted output string}
# EOF:{running milliseconds} 结束标记
# DBG:{debug string}
# LNO:{line number} # 从0开始

# 使用方法
# python3 {__file__}.py
# >>> print("hello", end="")
# LNO:0
# DBG:  0 print("hello", end="")
# WRT:"hello"
# EOF:1

import contextlib
import linecache
import json
import os
import sys
import traceback
import time
from typing import Union, Any


def exec_code(code: str, globals) -> Union[Any, None]:
    try:
        ccode = compile(code, "<string>", "eval")
        _eval = True
    except SyntaxError:
        ccode = compile(code, "<string>", "exec")
        _eval = False

    if _eval:
        return eval(ccode, globals)
    exec(ccode, globals)


_file_contents = {}


def getline(filename: str, lineno: int) -> str:
    """
    Args:
        lineno starts from 0

    Note:
        linecache.getline starts from 1
    """
    if os.path.isfile(filename):
        return linecache.getline(filename, lineno + 1)
    if filename == "<string>":
        lines = _file_contents[filename].splitlines()
        if lineno < len(lines):
            return lines[lineno]
        return ''


def gen_tracefunc(trace_filename: str, sys_stdout):
    """
    Ref: http://www.dalkescientific.com/writings/diary/archive/2005/04/20/tracing_python_code.html
    """

    def _trace(frame, event, arg):
        if event == "line":
            lineno = frame.f_lineno - 1  # set lineno starts from 0
            filename = frame.f_globals.get("__file__")

            if filename == trace_filename:
                line = getline(filename, lineno).rstrip()

                sys_stdout.write("LNO:{}\n".format(lineno))
                sys_stdout.write(f"DBG:{lineno:3d} {line}\n")
                sys_stdout.flush()
                # time.sleep(.5)

        return _trace

    return _trace


class QuitError(Exception):
    """ quit for this program """


@contextlib.contextmanager
def mock_stdout_stderr(prefix="WRT:"):
    _stdout = sys.stdout
    _stderr = sys.stderr
    try:

        class MockStdout:
            def write(self, data: str):
                try:
                    if data != "":
                        _stdout.write(prefix + json.dumps(data) + "\n")
                        _stdout.flush()
                except Exception as e:
                    raise QuitError("Output exception", str(e))

        sys.stdout = sys.stderr = MockStdout()
        yield _stdout, _stderr  # lambda s: _stdout.write(s+"\n")
    finally:
        sys.stdout = _stdout
        sys.stderr = _stderr


def stdin_readline():
    try:
        line = sys.stdin.readline().rstrip()
        if line.startswith("\""):
            line = json.loads(line)
        _file_contents["<string>"] = line
        # print(repr(line))
        return line
    except Exception as e:
        raise QuitError("readline", str(e))


def main():
    sigint_twice = False
    _globals = {
        "__file__": "<string>",
        "__name__": "__main__",
        "os": os,
        "sys": sys,
        "time": time,
        "json": json,
    }

    with mock_stdout_stderr() as (stdout, stderr):
        # preload
        import uiautomator2
        _globals['uiautomator2'] = uiautomator2

        sys.settrace(gen_tracefunc("<string>", stdout))
        stdout.write("DBG:Python (pid: {})\n".format(os.getpid()))
        while True:
            start = None

            try:
                # Read exec-code from stdin
                stderr.write(">>> ")
                stderr.flush()
                line = stdin_readline()

                start = time.time()
                sigint_twice = False

                ret = exec_code(line, _globals)
                if ret is not None:
                    print(ret)
            except KeyboardInterrupt:
                # Cancel running
                if sigint_twice:
                    break
                sigint_twice = True
                if start:
                    stdout.write(
                        "WRT:" +
                        json.dumps(">>> Catch Signal KeyboardInterrupt\n"))
                    stdout.write("\n")
                # stdout.write("INFO:KeyboardInterrupt catched, twice quit\n")
            except QuitError as e:
                stdout.write("DBG:{!r}".format(e))
                # Read error from stdin
                stdout.write("QUIT\n")
                break
            except:
                # Show traceback
                # https://docs.python.org/3/library/traceback.html
                flines = traceback.format_exc().splitlines(keepends=True)
                print(flines[0] +
                      "".join(flines[5:]).rstrip())  # ignore top 2 stack-frame
            finally:
                # Code block finished running
                millis = 0 if start is None else (time.time() - start) * 1000
                stdout.write("EOF:{}\n".format(int(millis)))
                stdout.flush()


if __name__ == "__main__":
    main()



# coding: utf-8
#
# rpc python server
# example rpc client
#
# import xmlrpc.client
# s = xmlrpc.client.ServerProxy("http://localhost:17320")
# s.run_python_code("a = 1")
# print(s.run_python_code('print(a)))
#
## output: 1

import argparse
import io
import os
import signal
import sys
import time
import traceback
from xmlrpc.server import SimpleXMLRPCRequestHandler, SimpleXMLRPCServer

import uiautomator2

_globals = {
    "sys": "sys",
    "time": time,
    "uiautomator2": uiautomator2,
    "u2": uiautomator2
}


class StringBuffer():
    def __init__(self):
        self.encoding = 'utf-8'
        self.buf = io.BytesIO()

    def write(self, data):
        if isinstance(data, str):
            data = data.encode(self.encoding)
        self.buf.write(data)

    def getvalue(self):
        return self.buf.getvalue().decode(self.encoding)


class FakeOutput(object):
    def __init__(self):
        self._fakeout = StringBuffer()  #io.StringIO()

    @property
    def output(self):
        return self._fakeout.getvalue()

    def __enter__(self):
        self.sys_stdout = sys.stdout
        self.sys_stderr = sys.stderr
        sys.stdout = self._fakeout
        sys.stderr = self._fakeout
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        sys.stdout = self.sys_stdout
        sys.stderr = self.sys_stderr


__cached_devices = {}


def _connect(device_id: str):
    """
    Returns:
        (u2.Device, fixed-device-id)
    """
    platform, uri = device_id.split(":", maxsplit=1)
    if platform == "android":
        d = uiautomator2.connect(uri)
        _id = "android:" + d.serial
        return d, _id
    elif platform == "ios":
        import wda
        d = wda.Client(uri)
        return d.session(), device_id
    else:
        raise RuntimeError("unknown platform", platform)


def get_device(device_id: str):
    """
    Returns:
        u2.Device
    """
    d = __cached_devices.get(device_id)
    if d:
        return d
    d, _id = _connect(device_id)
    if device_id.startswith("android:"):
        __cached_devices[_id] = d
    return d


def run_python_code(device_id: str, code: str):
    is_eval = True
    compiled_code = None
    try:
        compiled_code = compile(code, "<string>", "eval")
    except SyntaxError:
        is_eval = False
        compiled_code = compile(code, "<string>", "exec")

    with FakeOutput() as c:
        try:
            _globals.update({"d": get_device(device_id)})
            if is_eval:
                ret = eval(code, _globals)
                print(">>> " + repr(ret))
            else:
                exec(compiled_code, _globals)
        except:
            traceback.print_exc()

    return c.output


def main():
    parser = argparse.ArgumentParser(
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument("-p",
                        "--port",
                        type=int,
                        default=8000,
                        help="listen port")
    args = parser.parse_args()

    port = args.port
    server = SimpleXMLRPCServer(("0.0.0.0", port), allow_none=True)
    server.register_function(run_python_code)
    server.register_function(lambda: 'pong', 'ping')
    server.register_function(lambda x, y: x + y, 'add')
    server.register_function(lambda: os.kill(os.getpid(), signal.SIGTERM),
                             'quit')
    # server.register_function(_connect, "connect")
    server.register_multicall_functions()

    print(f'Serving XML-RPC on localhost port {port}')
    try:
        server.serve_forever()
    except KeyboardInterrupt:
        print("\nKeyboard interrupt received, exiting.")


if __name__ == "__main__":
    main()
